#include "LList.cpp"
#include <math.h>
#include <time.h>
#include <fstream.h>

struct particle
{
  particle() {};
  particle(float a, float b, float c) {x=a;y=b;z=c;}
  float x, y, z;
};

struct drawrule
{
  int when; // adjacency count to use this rule for
  char type; // s for sphere, c for cylinder, t for triangle
  float radius; // only defined for spheres and cylinders
  int order[3]; // order of particles to draw in -- 0 is the glider
  char style[255]; // style to draw in
};

#include "povout.cpp"

LList<particle> structure;
LList<drawrule> drawrules;

LList<particle> * getAdja(particle a, int min, int max, float t, bool & crowd);
float distance(particle a, particle b);
float disto(particle a);
void draw(povout & file, drawrule rule, LList<particle> & pop, particle glider);
particle getParticle(int number, LList<particle> & pop, particle glider);

void main(int argc, char * argv[])
{
  // ******** Initialize variables to empty ********
  int num_particles = -1;
  float release_radius = -1, kill_offset = -1;
  int min_adja = -1, max_adja = -1;
  float threshold = -1, velocity = -1;
  float max_turn = -1;
  int grad = 10000;
  int start_facing = -1;
  char out_filename[255] = {0};
  particle camera(10,10,10), focus(0,0,0), light(10,10,10);
  char lightstyle[255] = {0};
  char bgstyle[255] = {0};
  int randomseed = -1;
  // ***********************************************

  // ********* Parse *******************************
  char varfile[255] = {0};
  if (argc < 2) strcpy(varfile, "default.dla");
  if (argc > 1) strcpy(varfile, argv[1]);
  ifstream fin(varfile);
  char token[255] = {0};
  bool quit = 0;
  while (!quit)
  {
    fin >> token;
    if (strcmp(token, "rule") == 0)
    {
      drawrule temp;
      char item[255];
      fin >> temp.when;
      fin >> temp.type;
      fin >> temp.radius;
      int count = 0;
      if (temp.type == 's') count = 1;
      else if (temp.type == 'c') count = 2;
      else if (temp.type == 't') count = 3;
      char num[10];
      for (int i = 0; i < count; i++) fin >> temp.order[i];
      fin >> item;
      for (int i = 0; i < strlen(item); i++) if (item[i] == '_') item[i] = ' ';
      strcpy(temp.style,item);
      drawrules.insert(temp,0);
    }
    else if (strcmp(token, "outfile") == 0)
    {
      char item[255];
      fin >> item;
      strcpy(out_filename, item);
    }
    else if (strcmp(token, "randomize") == 0) fin >> randomseed;
    else if (strcmp(token, "do") == 0) fin >> num_particles;
    else if (strcmp(token, "releaseoffset") == 0) fin >> release_radius;
    else if (strcmp(token, "killoffset") == 0) fin >> kill_offset;
    else if (strcmp(token, "minadja") == 0) fin >> min_adja;
    else if (strcmp(token, "maxadja") == 0) fin >> max_adja;
    else if (strcmp(token, "threshold") == 0) fin >> threshold;
    else if (strcmp(token, "velocity") == 0) fin >> velocity;
    else if (strcmp(token, "turn") == 0) fin >> max_turn;
    else if (strcmp(token, "facing") == 0) fin >> start_facing;
    else if (strcmp(token, "randomize") == 0) fin >> randomseed;
    else if (strcmp(token, "camera") == 0) fin >> camera.x >> camera.y >> camera.z;
    else if (strcmp(token, "focus") == 0) fin >> focus.x >> focus.y >> focus.z;
    else if (strcmp(token, "light") == 0) fin >> light.x >> light.y >> light.z;
    else if (strcmp(token, "background") == 0)
    {
      char item[255];
      fin >> item;
      for (int i = 0; i < strlen(item); i++) if (item[i] == '_') item[i] = ' ';
      strcpy(bgstyle,item);
    }
    else if (strcmp(token, "lightstyle") == 0)
    {
      char item[255];
      fin >> item;
      for (int i = 0; i < strlen(item); i++) if (item[i] == '_') item[i] = ' ';
      strcpy(lightstyle,item);
    }
    else if (strcmp(token, "seed") == 0)
    {
      particle seed;
      fin >> seed.x >> seed.y >> seed.z;
      structure.insert(seed, 0);
    }
    else quit = 1;
  }
  fin.close();
  // ***********************************************
  cout << "Parse complete.  Beginning construction." << endl;
  // ******** Set variables to defaults ************
  if (num_particles <= 0) num_particles = 100;
  if (release_radius <= 1) release_radius = 2;
  if (kill_offset <= (release_radius + 1)) kill_offset = 2;
  if (min_adja < 1) min_adja = 1;
  if (max_adja < min_adja) max_adja = min_adja;
  if (threshold <= 0) threshold = .05;
  if (velocity <= 0) velocity = .5;
  if (max_turn <= 0) max_turn = 6.28;
  if (start_facing = -1) start_facing = 1;
  if (strlen(out_filename) == 0) strcpy(out_filename,"output.pov");
  if (strlen(lightstyle) == 0) strcpy(lightstyle,"color White");
  if (strlen(bgstyle) == 0) strcpy(bgstyle,"color White");
  if (randomseed < 0) randomseed = 0;
  // ***********************************************

  if (randomseed == 0) srand(time(0));
  else srand(randomseed);
  povout scene(out_filename);
  scene.setup(camera,focus,light,lightstyle,bgstyle);

  int counter = 0; int max_rad = 0;
  float a, b;
  while (counter < num_particles)
  {
    bool crowded = 0;
    a = (rand() % int(6.28 * grad)) / float(grad);
    b = (rand() % int(6.28 * grad)) / float(grad);
    particle glider((max_rad + release_radius) * sin(a), (max_rad + release_radius) * cos(a), (max_rad + release_radius) * sin(b));
    LList<particle> * adjacent = NULL;
    if (start_facing) { a = -a; b = -b; }
    do
    {
      if (adjacent != NULL) delete adjacent;
      adjacent = getAdja(glider, min_adja, max_adja, threshold, crowded);
      if (adjacent == NULL)
      {
        a = ((rand()%int(max_turn * grad))-(max_turn * grad / 2))/ float(grad);
        b = ((rand()%int(max_turn * grad))-(max_turn * grad / 2))/ float(grad);
        glider.x += velocity * sin(a);
        glider.y += velocity * cos(a);
        glider.z += velocity * sin(b);
      }
      if ((disto(glider) > (max_rad + release_radius + kill_offset))||crowded)
      {
        crowded = 0;
        a = (rand()% int(6.28 * grad)) / float(grad);
        b = (rand()% int(6.28 * grad)) / float(grad);
        glider.x = (max_rad + release_radius) * sin(a);
        glider.y = (max_rad + release_radius) * cos(a);
        glider.z = (max_rad + release_radius) * sin(b);
      }
    } while (adjacent == NULL);
    structure.insert(glider,0);
    if (disto(glider) > max_rad) max_rad = disto(glider);
    counter++;
    cout << counter << endl;
    for (LListIter<drawrule> arule(drawrules); !arule.atEnd(); arule.step())
      if (adjacent->size == (arule.item()).when)
        draw(scene, arule.item(), *adjacent, glider);
  }
}

LList<particle> * getAdja(particle a, int min, int max, float t, bool & crowd)
{
  LList<particle> * temp = NULL;
  int counter = 0;
  float dist;
  for (LListIter<particle> look(structure); !look.atEnd(); look.step())
  {
    dist = distance(a, look.item());
    if((dist <= (1.0 + t)) && (dist >= (1.0 - t)))
    {
      if (temp == NULL) temp = new LList<particle>;
      counter++;
      if (counter > max)
      {
        crowd = 1; return NULL;
      }
      temp->insert(look.item(), 0);
    }
  }
  if (counter < min) return NULL;
  else
  { crowd = 0; return temp; }
}

float distance(particle a, particle b)
{
  return sqrt(pow(a.x-b.x,2)+pow(a.y-b.y,2)+pow(a.z-b.z,2));
}

float disto(particle a)
{
  return sqrt(pow(a.x,2)+pow(a.y,2)+pow(a.z,2));
}

void draw(povout & file, drawrule rule, LList<particle> & pop, particle glider)
{
  if(rule.type == 's') // is a sphere
  {
    particle a = getParticle(rule.order[0], pop, glider);
    file.sphere(a, rule.radius, rule.style);
  }
  else if(rule.type == 'c') // is a cylinder
  {
    particle a = getParticle(rule.order[0], pop, glider);
    particle b = getParticle(rule.order[1], pop, glider);
    file.cylinder(a, b, rule.radius, rule.style);
  }
  else if(rule.type == 't') // is a triangle
  {
    particle a = getParticle(rule.order[0], pop, glider);
    particle b = getParticle(rule.order[1], pop, glider);
    particle c = getParticle(rule.order[2], pop, glider);
    file.triangle(a, b, c, rule.style);
  }
}

particle getParticle(int number, LList<particle> & pop, particle glider)
{
  if (number == 0) return glider;
  else
  {
    LListIter<particle> scan(pop);
    for (int i = 1; i < number; i++) if(!scan.atEnd()) scan.step();
    return scan.item();
  }
}
